/*
 * qarva-ts-indexer.cpp
 *
 *  Created on: Apr 18, 2015
 *      Author: root
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "../qcommon/Includes/qstdinc.h"
#include "../qcommon/Includes/qlogger.h"
#include "../qcommon/Includes/qsock.h"
#include "../qcommon/Includes/qsync.h"
#include "../qcommon/Includes/qtime.h"
#include "../qcommon/Includes/qdvbts.h"

#define QTSI_TAG 0x49535451
#define QPCRBASEMOD 0x200000000LL
#define QPCREXTEMOD 300LL


QLOG_REGISTER_SOURCE

struct QPCR
{
	QPCR()
	{
		nBase = (uint64_t)-1LL;
		nExt = 0;
	}

	inline bool operator >(const QPCR& nSecond)
	{
		if (nBase > nSecond.nBase) return true;
		if (nBase < nSecond.nBase) return false;
		if (nBase == nSecond.nBase) return nExt > nSecond.nExt;
		return false;
	}

	inline bool operator ==(const QPCR& nSecond)
	{
		if (nBase != nSecond.nBase) return false;
		if (nExt != nSecond.nExt) return false;
		return true;
	}

	inline QPCR operator -(const QPCR& nSecond)
	{
		QPCR nRetVal;

		if (nBase >= nSecond.nBase)
		{
			nRetVal.nBase = nBase - nSecond.nBase;
		}
		else
		{
			nRetVal.nBase = QPCRBASEMOD + nBase - nSecond.nBase;
		}

		if (nExt < nSecond.nExt)
		{
			if (nRetVal.nBase > 0)
			{
				nRetVal.nBase--;
			}
			else
			{
				nRetVal.nBase = (QPCRBASEMOD - 1LL);
			}

			nRetVal.nExt = QPCREXTEMOD + nExt - nSecond.nExt;
		}
		else
		{
			nRetVal.nExt = nExt - nSecond.nExt;
		}

		return nRetVal;
	}

	inline QPCR operator +(const QPCR& nSecond)
	{
		QPCR nRetVal;

		nRetVal.nBase = nBase + nSecond.nBase;
		nRetVal.nExt = nExt + nSecond.nExt;
		if (nRetVal.nExt > (QPCREXTEMOD - 1LL))
		{
			nRetVal.nExt = nRetVal.nExt % QPCREXTEMOD;
			nRetVal.nBase++;
		}

		if (nRetVal.nBase >= QPCRBASEMOD)
		{
			nRetVal.nBase = nRetVal.nBase % QPCRBASEMOD;
		}

		return nRetVal;
	}

	inline uint64_t Value() { return nBase * QPCREXTEMOD + nExt; }

	uint64_t	nBase;
	uint64_t	nExt;
};


struct QSendInfo
{
	uint64_t		nStart;			// Start position in file
	uint64_t		nSize;			// Size of this part in file
	QPCR			nBase;			// Diff from first PCR, base of send
	QPCR			nExtent;		// Extent of this part
};
typedef std::vector<QSendInfo> QSendInfos;


void PrintUsage()
{
	QLogInfo(0, "Use: qarva-ts-streamer [parameters]");
	QLogInfo(0, "    parameters:");
	QLogInfo(0, "        -daddr 'multicast address' (required)");
	QLogInfo(0, "        -dport 'multicast port' (required)");
	QLogInfo(0, "        -dif 'destination interface' (any interface by default)");
	QLogInfo(0, "        -v output some detailed debug info");
	QLogInfo(0, "        -r stream as RTP");
	QLogInfo(0, "        -i 'input file'");
	QLogInfo(0, "        -si save indices to 'input file'.idx (and finish)");
	QLogInfo(0, "        -li try to load saved indices from 'input file'.idx");
	QLogInfo(0, "        -lf also create log file, by default - only screen output is done");
	QLogInfo(0, "        -ss supress screen output, by default - screen output is enabled");
}

static int XCreateSocket(const char* pDstIF, const char* pDstIP, uint16_t nDstPort, sockaddr_in&	m_addr)
{
	int nSock = -1;

	QLogInfo(0, "%s / %s / %d", pDstIF, pDstIP, nDstPort);

	nSock = socket ( PF_INET, SOCK_DGRAM, 0 );
	if (nSock == -1)
	{
		QLogError(0, "Error (%d) creating socket", errno);
		return -1;
	}

	/* Bind socket */
	sockaddr_in addr;
	addr.sin_family =PF_INET;
	addr.sin_addr.s_addr = inet_addr(pDstIF);
	addr.sin_port = 0;
	if (bind ( nSock, ( sockaddr * ) &addr, sizeof ( addr ) ))
	{
		QLogError(0, "Error (%d) binding socket to interface (%s)", errno, pDstIF);
		close(nSock);
		return -1;
	}

	int ttl = 100;
	if (setsockopt ( nSock, IPPROTO_IP, IP_MULTICAST_TTL, (const char*)&ttl, sizeof ( ttl ) ))
	{
		QLogError(0, "Error (%d) setting TTL on socket", errno);
		close(nSock);
		return -1;
	}

	/* Setup multicast ADDR */
	m_addr.sin_family = PF_INET;
	m_addr.sin_addr.s_addr = inet_addr(pDstIP);
	m_addr.sin_port = htons(nDstPort);

	return nSock;
}

int main(int argc, char* argv[])
{
	/* Vars */
	uint64_t nFileSize = 0;
	int nInFile = -1;
	FILE* pIDX = NULL;
	uint8_t* pDData = NULL;
	uint8_t refdata[188 * 7];

	QLogInfo(0, "=========================================");
	QLogInfo(0, "Qarva Aqua TS file streamer %s", QGetVersion());
	QLogInfo(0, "=========================================");

	if (argc < 2)
	{
abortparse:
		PrintUsage();
		return 0;
	}

	char sDstAddr[128];
	char sDstPort[128];
	char sDstIF[128];
	bool bVerbose = false;
	char sInputFile[128];
	char sIdxFile[128];
	bool bLogFile = false;
	bool bLogScreen = true;
	bool bRTP = false;
	bool bLoadIndex = false;
	bool bSaveIndex = false;
	uint16_t fSeqNo = (uint16_t)time(NULL);

	/* Init some params */
	memset(sDstAddr, 0, sizeof(sDstAddr));
	memset(sDstPort, 0, sizeof(sDstPort));
	strcpy(sDstIF, "0.0.0.0");
	memset(sInputFile, 0, sizeof(sInputFile));
	memset(sIdxFile, 0, sizeof(sIdxFile));

	/* Handle parameters */
	for (int i=1; i<argc; i++)
	{
		if (strcmp(argv[i], "-daddr") == 0)
		{
			i++;
			if (i < argc)
			{
				strcpy(sDstAddr, argv[i]);
			}
			else
			{
				QLogError(0, "Absent parameter for -daddr");
				goto abortparse;
			}
		}
		else if (strcmp(argv[i], "-dport") == 0)
		{
			i++;
			if (i < argc)
			{
				strcpy(sDstPort, argv[i]);
			}
			else
			{
				QLogError(0, "Absent parameter for -dport");
				goto abortparse;
			}
		}
		else if (strcmp(argv[i], "-dif") == 0)
		{
			i++;
			if (i < argc)
			{
				strcpy(sDstIF, argv[i]);
			}
			else
			{
				QLogError(0, "Absent parameter for -dif");
				goto abortparse;
			}
		}
		else if (strcmp(argv[i], "-v") == 0)
		{
			bVerbose = true;
		}
		else if (strcmp(argv[i], "-r") == 0)
		{
			bRTP = true;
		}
		else if (strcmp(argv[i], "-si") == 0)
		{
			bSaveIndex = true;
		}
		else if (strcmp(argv[i], "-li") == 0)
		{
			bLoadIndex = true;
		}
		else if (strcmp(argv[i], "-i") == 0)
		{
			i++;
			if (i < argc)
			{
				strcpy(sInputFile, argv[i]);
				strcpy(sIdxFile, sInputFile);
				strcat(sIdxFile, ".idx");
			}
			else
			{
				QLogError(0, "Absent parameter for -i");
				goto abortparse;
			}
		}
		else if (strcmp(argv[i], "-lf") == 0)
		{
			bLogFile = true;
		}
		else if (strcmp(argv[i], "-ss") == 0)
		{
			bLogScreen = false;
		}
		else
		{
			QLogError(0, "Invalid parameter (%s)", argv[i]);
			goto abortparse;
		}
	}

	/* Verify params */
	if (strlen(sDstAddr) == 0) goto abortparse;
	if (strlen(sDstPort) == 0) goto abortparse;
	if (strlen(sInputFile) == 0) goto abortparse;

	/* Init logger */
	QLogInit(bLogScreen, bLogFile, "qarva-ts-streamer");

	/* Create socket */
	sockaddr_in	addrX;
	socklen_t addrlen = sizeof(addrX);
	int nSocket = XCreateSocket(sDstIF, sDstAddr, atoi(sDstPort), addrX);
	if (nSocket == -1) return 0;

	/* Step 0 - open files */
	nInFile = open(sInputFile, O_RDWR);

	if (nInFile == -1)
	{
		QLogError(0, "Could not open input file (%s)", sInputFile);
		return 0;
	}

	/* Step 1 - detect file size */
	nFileSize = lseek(nInFile, 0L, SEEK_END);
	if (nFileSize == (uint64_t)-1)
	{
		QLogError(0, "Could not seek to the end of file (%d)", errno);
		return 0;
	}

	QLogInfo(0, "File size is: %llu", nFileSize);
	if (lseek(nInFile, 0L, SEEK_SET))
	{
		QLogError(0, "Could not seek to the beginning of file (%d)", errno);
		return 0;
	}

	if (nFileSize % 188)
	{
		nFileSize = nFileSize / 188;
		nFileSize = nFileSize * 188;
		QLogError(0, "File size if not multiply of 188, not a TS file? Will be truncated to (%llu)", nFileSize);
	}

	/* Step 1 - read file into RAM */
	QLogInfo(0, "Allocating memory mapping for file...");
	pDData = (uint8_t*)mmap(NULL, nFileSize, PROT_READ | PROT_WRITE, MAP_SHARED, nInFile, 0);
	if (pDData == MAP_FAILED)
	{
		QLogInfo(0, "File mapping failed, error : %d", errno);
		return 0;
	}

	uint64_t nDone = 0;
	uint64_t nPCT = 10;
	uint64_t nTest = 0;
	QSendInfos infos;

	QLogInfo(0, "Processing PCR info...");
	if (bVerbose) QLogInfo(0, "=========================================");

	nDone = 0;
	nPCT = 10;
	nTest = 0;

	uint64_t nSeconds = 0;

	if (bLoadIndex)
	{
		pIDX = fopen(sIdxFile, "r+b");
		if (pIDX == NULL)
		{
			QLogInfo(0, "Error opening index file...");
			return 0;
		}
		uint32_t nCount = 0;
		if (fread(&nCount, sizeof(nCount), 1, pIDX) != 1)
		{
			QLogInfo(0, "Can not read tag from index file...");
			return 0;
		}

		if (nCount != QTSI_TAG)
		{
			QLogInfo(0, "Invalid index file...");
			return 0;
		}

		if (fread(&nSeconds, sizeof(nSeconds), 1, pIDX) != 1)
		{
			QLogInfo(0, "Can not read index count from index file...");
			return 0;
		}

		if (fread(&nCount, sizeof(nCount), 1, pIDX) != 1)
		{
			QLogInfo(0, "Can not read index count from index file...");
			return 0;
		}

		for (uint32_t i=0; i<nCount; i++)
		{
			QSendInfo info;

			if (fread(&info, sizeof(info), 1, pIDX) != 1)
			{
				QLogInfo(0, "Can not read info (%u) of (%u) from index file...", i, nCount);
				return 0;
			}

			infos.push_back(info);
		}

		if (fread(&nCount, sizeof(nCount), 1, pIDX) != 1)
		{
			QLogInfo(0, "Can not read final tag from index file...");
			return 0;
		}

		if (nCount != QTSI_TAG)
		{
			QLogInfo(0, "Incomplete index file...");
			return 0;
		}

		fclose(pIDX);
		pIDX = NULL;
	}
	else
	{
		QPCR nFirstPCR;
		uint64_t nFirstPCRPos = (uint64_t)-1LL;

		QPCR nLastPCR;
		uint64_t nLastPCRPos = (uint64_t)-1LL;

		int nPCRCount = 0;

		while(nDone < nFileSize)
		{
			QDVBTSPacket ts;
			QPCR nPCR;
			QPCR nDiff;

			if (!QDVBParseTS(&pDData[nDone], &ts))
			{
				QLogError(0, "Unable to parse TS at position (%llu)", nDone);
				return 0;
			}
			else
			{
				if (ts.PCR_flag)
				{
					nPCR.nBase = ts.program_clock_reference_base;
					nPCR.nExt = ts.program_clock_reference_extension;

					nPCRCount++;
					if (nFirstPCR.nBase == (uint64_t)-1LL)
					{
						nFirstPCR = nPCR;
						nFirstPCRPos = nDone;
						if (bVerbose) QLogInfo(0, "First PCR (%llu) found at position (%llu), PID(%u)...", nFirstPCR.Value(), nDone, ts.PID);
					}

					if (nLastPCR.nBase != (uint64_t)-1LL)
					{
						if (nLastPCR > nPCR)
						{
							nDiff = nLastPCR - nPCR;
							if (bVerbose) QLogError(0, "PCR (%llu) found at position (%llu) is lesser than previous PCR(%llu) at position (%llu), PID(%u), possible turn-around (%llu)...", nPCR.Value(), nDone, nLastPCR.Value(), nLastPCRPos, ts.PID, nDiff.Value());
						}
						else
						{
							nDiff = nPCR - nLastPCR;
						}

						/* Add send info */
						QSendInfo info;
						info.nBase = nLastPCR - nFirstPCR;

						info.nExtent = nDiff;
						info.nStart = nLastPCRPos;
						info.nSize = nDone - nLastPCRPos;
						infos.push_back(info);
					}

					/* Note 27 000 000 is 1 sec */
					/* Or in nBase units 90000 */
					/* No less than 0.01, no more than 0.5 */

					if (nDiff.nBase != (uint64_t)-1LL)
					{
						if (nDiff.Value() == 0)
						{
							/* Special case, check that we have no difference more that 1 UDP */
							if ((nDone - nLastPCRPos) > (188 * 7))
							{
								if (bVerbose) QLogWarning(0, "The same PCR (%llu) is found at position (%llu),  previous PCR(%llu) is at position (%llu), PID(%u), but distance is big (%llu)...", 
									nPCR.Value(), nDone, nLastPCR.Value(), nLastPCRPos, ts.PID, (uint64_t)(nDone - nLastPCRPos));
							}
							nPCRCount--;
						}
						else if (nDiff.nBase < 900)
						{
							if (bVerbose) QLogWarning(0, "Too small diff(%llu) from previous PCR(%llu) is found at position (%llu),  previous PCR(%llu) at position (%llu), PID(%u), distance is (%llu)...", 
								nDiff.Value(), nPCR.Value(), nDone, nLastPCR.Value(), nLastPCRPos, ts.PID, (uint64_t)(nDone - nLastPCRPos));
						}
						else if (nDiff.nBase > 45000)
						{
							if (bVerbose) QLogError(0, "Too big diff(%llu) from previous PCR(%llu) is found at position (%llu),  previous PCR(%llu) at position (%llu), PID(%u), distance is (%llu)...", 
								nDiff, nPCR.Value(), nDone, nLastPCR.Value(), nLastPCRPos, ts.PID, (uint64_t)(nDone - nLastPCRPos));
							return 0;
						}
					}

					nLastPCR = nPCR;
					nLastPCRPos = nDone;
				}
			}

			nDone += 188;
			nTest = nDone * 100 / nFileSize;
			if (nTest >= nPCT)
			{
				if (bVerbose) QLogInfo(0, "Parsed %llu%c...", nPCT, '%');
				nPCT += 10;
			}
		}

		/* Add first slack */
		if (nFirstPCRPos != 0)
		{
			QSendInfo info;

			// First part will burst
			info.nBase = nFirstPCR;
			info.nExtent.nBase = 0;
			info.nExtent.nExt = 0;

			info.nStart = 0;
			info.nSize = nFirstPCRPos;
			infos.insert(infos.begin(), info);
		}

		if (nFileSize - nLastPCRPos)
		{
			QSendInfo info;

			// Last part will burst
			info.nBase = nLastPCR;
			info.nExtent.nBase = 0;
			info.nExtent.nExt = 0;

			info.nStart = nLastPCRPos;
			info.nSize = nFileSize - nLastPCRPos;
			infos.push_back(info);
		}

		if (bVerbose) QLogInfo(0, "Starting PCR(%llu) found at position (%llu), ending  PCR(%llu) found at position (%llu), ending slack (%llu)...", 
			nFirstPCR.Value(), nFirstPCRPos, nLastPCR.Value(), nLastPCRPos, (uint64_t)(nFileSize - nLastPCRPos));
		if (bVerbose) QLogInfo(0, "PCRs found in file (%d)...", nPCRCount);
		nSeconds = (nLastPCR - nFirstPCR).Value() / 27000000LL;

		/* Save index if requested */
		if (bSaveIndex)
		{
			pIDX = fopen(sIdxFile, "w+b");
			if (pIDX == NULL)
			{
				QLogInfo(0, "Error creating index file...");
				return 0;
			}
			uint32_t nCount = QTSI_TAG;
			if (fwrite(&nCount, sizeof(nCount), 1, pIDX) != 1)
			{
				QLogInfo(0, "Can not write tag to index file...");
				return 0;
			}

			if (fwrite(&nSeconds, sizeof(nSeconds), 1, pIDX) != 1)
			{
				QLogInfo(0, "Can not write seconds to index file...");
				return 0;
			}

			nCount = (uint32_t)infos.size();
			if (fwrite(&nCount, sizeof(nCount), 1, pIDX) != 1)
			{
				QLogInfo(0, "Can not write count to index file...");
				return 0;
			}

			for (uint32_t i=0; i<nCount; i++)
			{
				QSendInfo info = infos[i];

				if (fwrite(&info, sizeof(info), 1, pIDX) != 1)
				{
					QLogInfo(0, "Can not write info (%u) of (%u) to index file...", i, nCount);
					return 0;
				}
			}

			nCount = QTSI_TAG;
			if (fwrite(&nCount, sizeof(nCount), 1, pIDX) != 1)
			{
				QLogInfo(0, "Can not write final tag to index file...");
				return 0;
			}

			fflush(pIDX);
			fclose(pIDX);
			pIDX = NULL;
			return 0;
		}
	}

	/* Validate infos */
	if (bVerbose)
	{
		QPCR nSum;
		uint64_t nLen = 0;
		nSum.nBase = 0;
		nSum.nExt = 0;
		QSendInfo infof;
		for (int i=0; i<(int)infos.size(); i++)
		{
			QSendInfo& info = infos[i];
			if (info.nBase.nBase != (uint64_t)-1LL)
			{
				infof = info;
				nSum = nSum + info.nExtent;
			}
			nLen = nLen + info.nSize;
		}

		uint64_t nSecondsX = nSum.Value() / 27000000;
		QLogInfo(0, "Info calculated length is (%llu), calculated time %02d:%02d:%02d (or raw %d seconds)......", nLen, (nSecondsX / 60) / 60,  (nSecondsX / 60) % 60, nSecondsX % 60, nSecondsX);
		nSecondsX = infof.nBase.Value() / 27000000;
		QLogInfo(0, "Full info calculated time %02d:%02d:%02d (or raw %d seconds)......", (nSecondsX / 60) / 60,  (nSecondsX / 60) % 60, nSecondsX % 60, nSecondsX);
	}

	QLogInfo(0, "File length is %02d:%02d:%02d (or raw %d seconds)...", (nSeconds / 60) / 60,  (nSeconds / 60) % 60, nSeconds % 60, nSeconds);

	QLogInfo(0, "=========================================");
	QLogInfo(0, "Start streaming...");
	QLogInfo(0, "=========================================");

	/* Create reference buffer */
	memset(refdata, -1, 188 * 7);
	for (int i = 0; i < 7; i++)
	{
		uint8_t* pTmp = &refdata[188 * i];
		QDVBTSPacket ots;
		memset(&ots, 0, sizeof(ots));
		ots.sync_byte = 'G';
		ots.PID = 0x1FFF;
		ots.adaptation_field_control = 2;
		ots.continuity_counter = 0;
		ots.adaptation_field_length = 183;
		QDVBMakeTS(pTmp, &ots);
	}

	/* Step 3 - segment data with timestamps */
	while(true)
	{
		/* Base send time */
		timespec tmbase;
		if (clock_gettime(CLOCK_MONOTONIC_RAW, &tmbase))
		{
			QLogInfo(0, "Base timer error: (%d)", errno); exit(0);
		}

		timespec tmsend;
		uint8_t bufsendrtp[12 + 188 * 7];
		uint8_t* bufsend = &bufsendrtp[12];
		uint64_t nTotalSent = 0;
		uint32_t nStartTime = (uint32_t)time(NULL);
		uint32_t nLastTime = nStartTime;

		for (int i = 0; i < (int)infos.size(); i++)
		{
			QSendInfo& info = infos[i];
			int nToSend = 0;
			int nDone = 0;

			uint64_t nStartPCR = info.nBase.Value();
			uint64_t nDiffPCR = info.nExtent.Value();

			/* Form buffer */
			while (nDone < (int)info.nSize)
			{
				/* Prepare main data */
				memcpy(bufsend, refdata, 188 * 7);
				nToSend = std::min((int)(info.nSize - nDone), (int)(188 * 7));
				memcpy(bufsend, &pDData[info.nStart + nDone], nToSend);

				/* Ok, diff from start in 27 MHz */
				uint64_t nShift = nDiffPCR * (uint64_t)nDone / (uint64_t)info.nSize;
				uint64_t nSendPCR = nStartPCR + nShift;

				/* Now convert nSendPCR into timespec structure */
				uint64_t nSSec = nSendPCR / 27000000LL;
				uint64_t nSNSec = nSendPCR % 27000000LL;

				/* 27 000 000 is equal to 1 000 000 000 */
				nSNSec = nSNSec * 1000 / 27;
				tmsend.tv_sec = nSSec;
				tmsend.tv_nsec = nSNSec;

				//QLogInfo(0, "Send shift for part(%u:%u:%u) is (%u) sec (%u) nsec:", i, nDone, (int)info.nSize, (int)tmsend.tv_sec, (int)tmsend.tv_nsec);

				/* Add start timespec */
				tmsend.tv_sec += tmbase.tv_sec;
				tmsend.tv_nsec += tmbase.tv_nsec;
				if (tmsend.tv_nsec > 1000000000LL)
				{
					tmsend.tv_nsec = tmsend.tv_nsec % 1000000000LL;
					tmsend.tv_sec++;
				}

				/* Calculate precise timing, just send info */
				timespec tmnow;
				while (true)
				{
					if (clock_gettime(CLOCK_MONOTONIC_RAW, &tmnow))
					{
						QLogInfo(0, "Send timer error: (%d)", errno); exit(0);
					}

					/* Compare */
					if (tmnow.tv_sec > tmsend.tv_sec) break;
					if (tmnow.tv_sec == tmsend.tv_sec)
					{
						if (tmnow.tv_nsec > tmsend.tv_nsec) break;
					}
				}

				/* Prepare RTP header if required */
				if (bRTP)
				{
					/* Header */
					uint32_t rtpHdr = 0x80000000;
					rtpHdr |= (33 << 16);
					rtpHdr |= fSeqNo++; // sequence number
					*((uint32_t*)&bufsendrtp[0]) = htonl(rtpHdr);

					/* Timestamp */
					uint64_t nXPCR = tmsend.tv_sec * (uint64_t)10000000000LL;
					nXPCR = nXPCR + (uint64_t)tmsend.tv_nsec;
					nXPCR = nXPCR * (uint64_t)9LL / (uint64_t)1000000LL;
					*((uint32_t*)&bufsendrtp[4]) = htonl((uint32_t)nXPCR);

					/* Session */
					*((uint32_t*)&bufsendrtp[8]) = 0;
				}

				/* Check buffer for 'G' */
				for (int c = 0; c < 7; c++)
				{
					if (bufsend[c * 188] != 'G')
					{
						QLogError(0, "Invalid TS signature at packet (%d), actual packets (%d)(%d)(%d)", c, nToSend, nToSend / 188, nToSend % 188);
					}
				}

				/* Send data at last */
				int nSent = 0;
				if (bRTP)
				{
					nSent = (sendto(nSocket, bufsendrtp, nToSend + 12, MSG_NOSIGNAL, (const sockaddr*)&addrX, addrlen) - 12);
				}
				else
				{
					nSent = sendto(nSocket, bufsend, nToSend, MSG_NOSIGNAL, (const sockaddr*)&addrX, addrlen);
				}

				if (nSent != nToSend)
				{
					QLogError(0, "Invalid send size on socket (%d)", nSent);
					return 0;
				}
				nDone += nToSend;
				nTotalSent += (uint64_t)nToSend;
			}

			uint32_t nTimeNow = (uint32_t)time(NULL);
			if (nLastTime < nTimeNow)
			{
				QLogInfo(0, "(%u%%) Sent %llu bytes in %u secs", i * 100 / (int)infos.size(), nTotalSent, nTimeNow - nStartTime);
				nLastTime = nTimeNow;
			}
		}
	}

	return 0;
}
